/*********************************************************************
 * Purpose:     A program to apply a function to two sets of numbers
 *              x and y, then sum the results of applying the function
 *              to every possible combination of the elements of those
 *              two sets.
 *
 * Author:      Steven Alexander
 * Class :      CPS 615
 * Due   :      March 29, 2000
 *******************************************************************/

#include	<stdio.h>
#include	"mpi.h"
#include        <math.h>

#define tag   0

int randval(int low, int high); /*gets a random int*/

void main ( int argc, char *argv[] )
{
  int rank, nextrank, prevrank, size, i, j, k, m, num_procs;
  int total = 0;
  int collector[10];
  int sum_x = 0;
  int send_y[10];
  int recv_y[10];
  int x[10];
  int cap_n, n;

  MPI_Status status;

  MPI_Init(&argc, &argv);         /*start MPI*/


  MPI_Comm_rank(MPI_COMM_WORLD, &rank);       /*get the rank of this process*/
  MPI_Comm_size(MPI_COMM_WORLD, &num_procs);  /*get the number of processes*/
  
  cap_n = 24; /*make up a problem size (I used a convenient number)*/
  n = cap_n/num_procs;  /*problem size divided by number of processes*/

  for(i = 0; i < n; i++)
  {				/*fill up arrays with values*/
    x[i] = 1;
  }
  for(i = 0; i < n; i++)
  {
    send_y[i] = randval(1, 5);
    recv_y[i] = randval(1, 5);
  }      

  /* set up each process so it knows which
     process is next in the ring and which
     process is before it.  */

  if(rank == num_procs - 1) nextrank = 0;                                         

  else nextrank = (rank + 1);

  if(rank == 0) prevrank = (num_procs - 1);     

  else prevrank = rank - 1;

  /*do num_procs times so all y vals are used*/
  for(i = 0; i < num_procs; i++) 
  {

     /*pass y values around in a ring structure so that we can
       keep track of where they are going and where they are
       coming from. */     

     

     if((rank % 2) == 0) /* even ranks send first*/
     {
	for(j = 0; j < n; j++)
        {
          for(k = 0; k < n; k++)
          {
            sum_x = sum_x + (x[j] * x[j] + recv_y[k] * recv_y[k]);
            
          }
        }
        for(m = 0; m < n; m++)
        {
         send_y[m] = recv_y[m];
        }

       /*send this process' y values to the next process*/
       MPI_Ssend(send_y, n, MPI_INT, nextrank, tag, MPI_COMM_WORLD);

       /*receive the y values from the previous process*/
       MPI_Recv(recv_y, n, MPI_INT, prevrank, tag, MPI_COMM_WORLD, &status);

     }

     else  /*odd ranks receive first*/
     {
	for(j = 0; j < n; j++)
        {
          for(k = 0; k < n; k++)
          {
            sum_x = sum_x + (x[j] * x[j] + recv_y[k] * recv_y[k]);
          }
        }        

        for(m = 0; m < n; m++)
        {
         send_y[m] = recv_y[m];
        }

	/*receive the y values from the previous process*/
        MPI_Recv(recv_y, n, MPI_INT, prevrank, tag, MPI_COMM_WORLD, &status);

        /*send this process' y values to the next process*/
        MPI_Ssend(send_y, n, MPI_INT, nextrank, tag, MPI_COMM_WORLD);

      }    

  } 

  /*gather all the sums onto process 0 and add them up*/

  MPI_Gather(&sum_x, 1, MPI_INT, collector, 1, MPI_INT, 0, MPI_COMM_WORLD);
  if (rank == 0)
   {
    for(i = 0; i < num_procs; i++)
    {
     total = total + collector[i];
    }
    /*print the sum of all sums from all processes (total)*/
    printf("The total sum is %d", total); 
   }

  MPI_Finalize();

} /* end main */

int randval(int low, int high) /*makes a random int between high and low*/
{
  int val;
  val = rand()%10;
  return (val);
} /* end randval */
